#! /usr/bin/env python3
# -*- coding: utf-8; mode: python -*-
""" Implementation of Hidden Markov Models, simple case, for Homework 3 of the PGM course (Master MVA).

We already did:
- alpha/beta recursion (forward/backward propagation),
- likelihood (and log-likelihood) computations,
- EM algorithm (estimating the parameters pi, A, mus, Sigmas),
- Viterbi algorithm (MAP inference / max-product algorithm),
- display 2D clustering from argmax on marginal distribution or Viterbi (question 8),
- display t -> most likely q_t in {1,..,K} (questions 10 and 11),
- tweak the filtering to be permutation invariant (to be able to compare the different filtering question2 and question9),
- add this permutation invariance for Viterbi too.
- Experimental: automatic tuning of the best number of states (learn K, question 12).


This HMM.py program requires other programs:

- EM.py for the Estimation-Maximisation algorithm for the Mixture of Gaussian model,
- kmeans.py is required by EM.py, not by HMM.py.


Homework assignement was on http://www.di.ens.fr/~slacoste/teaching/MVA_GM/fall2015/HWK3_2015.pdf
With data from http://www.di.ens.fr/~slacoste/teaching/MVA_GM/fall2015/classification_data_HWK3.zip


- *Date:* Saturday 19 December 2015.
- *Author:* Lilian Besson and Valentin Brunck, for the MVA Master, (C) 2015-16.
- *Licence:* MIT Licence (http://lbesson.mit-license.org).
"""

from __future__ import print_function, division  # Python 2 compatibility

from os.path import join as ospathjoin
import numpy as np
import numpy.linalg as la

import matplotlib.pyplot as plt

# From our modules
# from kmeans import k_means  # Local import
import EM
from EM import EM_general  # Local import

try:
    from scipy.misc import logsumexp  # log trick (to reduce underflow risks)
except ImportError:
    def logsumexp(li):
        """ Our own log-sum-exp function (**naive**, not optimized)."""
        return np.log(np.sum(np.exp(li)))
    print("We defined our own logsumexp function, scipy.misc.logsumexp should be use instead!")

# Option for customizing some part of the code
doSaveThePlots = True  # True to save the figures
saveAlsoPDF    = False  # Save also to PDF (default is just PNG)


# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# %% Utility functions on Gaussian densities

def gaussian_density(xi, muj, Sigmaj, d=None):
    """ N(x_i | mu_j, Sigma_j) : density of a Gaussian of center muj and covariance matrix Sigmaj."""
    if d is None:
        d = np.shape(xi)[0]
    xc = xi - muj
    return np.exp(- (np.dot(xc.T, np.dot(la.inv(Sigmaj), xc))) / 2) / np.sqrt(la.det(Sigmaj) * (2*np.pi)**d)


def log_gaussian_density(xi, muj, Sigmaj, d=None):
    """ Log normale density for xi, from parameters mu_j Sigma_j (log N(x_i | mu_j, Sigma_j)):

    log density = -1/2 d log(2 pi) -1/2 x^T . S . x -1/2 log det(S)
    """
    if d is None:
        d = np.shape(xi)[0]
    xc = xi - muj  # xi centered
    log_px = - 0.5 * d * np.log(2*np.pi)  # -1/2 d log(2 pi)
    log_px -= 0.5 * np.dot(xc.T, np.dot(la.inv(Sigmaj), xc))  # -1/2 x^T . S . x
    log_px -= 0.5 * np.log(la.det(Sigmaj))  # -1/2 log det(S)
    return log_px


# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# %% 1. alpha/beta recursion
print("\n\n1. alpha/beta recursion for HMM :")
print("Reference :\n - http://www.di.ens.fr/~slacoste/teaching/MVA_GM/fall2015/lecture7.pdf#page=11,\n - https://en.wikipedia.org/wiki/Hidden_Markov_model\n")


def alpha_recursion(u, A, mus, Sigmas, pi_init):
    """ Alpha recursion for HMM, working only with log for improved numerical stability (no underflow risk).

    - Complexity: O(T K),
    - Return: log_alpha, vector of log values of alpha.
    """
    # 0. Creating variables
    T = np.shape(u)[0]    # Number of points and dimension
    K = np.shape(mus)[0]  # Number of clusters
    log_alpha = np.zeros((T, K))
    # 1. Initialization for alpha : starting from t=0 (forward)
    for q in range(K):
        mu_q = mus[q, :]
        sigma_q = Sigmas[q]
        log_p_u_t_knowing_q_t = log_gaussian_density(u[0, :], mu_q, sigma_q)
        # Virtual message for alpha[0]
        log_alpha[0, q] = log_p_u_t_knowing_q_t + np.log(pi_init[q])
    # 2. Loop for computing the alpha !
    for t in range(1, T):  # forward !
        ut = u[t, :]
        for q in range(K):
            mu_q = mus[q, :]
            sigma_q = Sigmas[q]
            # a. From the Gaussian density for state q : mu_q, sigma_q
            log_p_u_t_knowing_q_t = log_gaussian_density(ut, mu_q, sigma_q)
            # b. From the transition matrix A
            log_p_q_t_knowing_q_t_minus_1 = np.log(A[:, q])
            ai = log_p_q_t_knowing_q_t_minus_1 + log_alpha[t-1, :].T
            # Here we use the log-sum-exp trick ! XXX
            max_ai = np.max(ai)
            log_alpha[t, q] = log_p_u_t_knowing_q_t + max_ai + logsumexp(ai - max_ai)
    # 4. Done for log-alpha recursion
    return log_alpha


def beta_recursion(u, A, mus, Sigmas):
    """ Beta recursion for HMM, working only with log for improved numerical stability (no underflow risk).

    - Complexity: O(T K^2),
    - Return: log_beta, vector of log values of beta.
    """
    # 0. Creating variables
    T = np.shape(u)[0]    # Number of points and dimension
    K = np.shape(mus)[0]  # Number of clusters
    log_beta = np.zeros((T, K))
    # 1. Initialization for beta : starting from t=T (backward)
    # zeros, but just to show Beta_T(z_T) = 1
    # log_beta[T-1, :] = np.log(np.ones((K, 1)))
    # 2. Loop for computing the beta !
    for t in range(T-2, -1, -1):  # backward !
        u_t_plus_1 = u[t+1, :]
        for q in range(K):
            ai = np.zeros((K, 1))  # vector sum
            for q_t_plus_1 in range(K):
                # a. From the Gaussian density for state q_t_plus_1 : mu_k, sigma_k
                mu_q = mus[q_t_plus_1, :]
                sigma_q = Sigmas[q_t_plus_1]
                log_p_u_t_plus_1_knowing_q_t_plus_1 = log_gaussian_density(u_t_plus_1, mu_q, sigma_q)
                # b. From the transition matrix A
                log_p_q_t_plus_1_knowing_q_t = np.log(A[q, q_t_plus_1])

                # Now adding things up
                ai[q_t_plus_1] = log_p_q_t_plus_1_knowing_q_t + \
                    log_p_u_t_plus_1_knowing_q_t_plus_1 + \
                    log_beta[t+1, q_t_plus_1]
            # End of this vector sum
            # Here we use the log-sum-exp trick ! XXX
            max_ai = np.max(ai)
            log_beta[t, q] = max_ai + logsumexp(ai - max_ai)
    # 4. Done for log-beta recursion
    return log_beta


def alphabeta(u, A, mus, Sigmas, pi_init):
    """ alpha/beta recursion for HMM (forward-backward algorithm, aka sum-product), with the log-trick to reduce underflow risk.

    - Return: log_alpha, log_beta,
    - Reference:
      - https://en.wikipedia.org/wiki/Hidden_Markov_model,
      - Bishop, Pattern Recognition and Machine Learning, § 13.2.2 (page 618),
      - http://www.di.ens.fr/~slacoste/teaching/MVA_GM/fall2015/lecture7.pdf#page=11.
    """
    log_alpha = alpha_recursion(u, A, mus, Sigmas, pi_init)
    log_beta = beta_recursion(u, A, mus, Sigmas)
    return log_alpha, log_beta


print("Note that we also tried to implement the alpha/beta recursion WITHOUT log trick, to compare, and as expected we only get zeros (underflow risk is real)")

# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# %% 2. 4 plots, clusterization from a trivial HMM
print("\n\n2. 4 plots, clusterization from a trivial HMM :")


def compute_filtering(log_alpha, log_beta, K, t_max):
    """ Use the values of alpha and beta

    - Return: the marginal distribution of hidden states p(q_t | u1,..,uT),
    - Note: it is called *filtering* because this operation comes from denoising sound,
    - Reference: Bishop, Pattern Recognition and Machine Learning, § 13.2.2 (page 618).
    """
    filtering = np.zeros((t_max+1, K))
    for t in range(t_max+1):
        ai = log_alpha[t, :] + log_beta[t, :]
        max_ai = np.max(ai)
        log_normalization = max_ai + logsumexp(log_alpha[t, :] + log_beta[t, :] - max_ai)
        filtering[t, :] = np.exp(log_alpha[t, :] + log_beta[t, :] - log_normalization)
    return filtering


def permutation_invariance(filtering, verb=False):
    """ Tweak the filtering to be permutation invariant (to be able to compare the different filtering question2 and question9).

    - Basically, it shuffle the states with a good permutation of {1,..,K} in order to always have the same order.
    - Quite inefficiently written, but it works...
    """
    n, K = np.shape(filtering)
    permutation = {k: k for k in range(K)}
    if verb: print("permutation =", permutation)  # DEBUG
    q = 0
    seen = set()
    for t in range(n):
        k = np.argmax(filtering[t, :])
        if verb: print("t = {}, k = {}, seen = {}".format(t, k, seen))  # DEBUG
        if k not in seen:
            if verb: print("  k not seen!")  # DEBUG
            permutation[q] = k
            if verb: print("  permutation =", permutation)  # DEBUG
            seen.add(k)
            if verb: print("  k =", k, "added to seen...")  # DEBUG
            q += 1
            if verb: print("  q =", q)  # DEBUG
        if q >= K:
            break
    new_filtering = np.zeros_like(filtering)
    for k in range(K):
        new_filtering[:, k] = filtering[:, permutation[k]]
    return new_filtering,permutation


def question2(u, K=4, t_max=100, save=doSaveThePlots):
    """ Question 2 : K plots, clusterization from a trivial HMM."""
    # Using the means and covariances obtained from MoG from HWK2
    mus, Sigmas, _ = EM_general(u, K)
    # Initial pi (uniform), A estimations :
    # Aii = 1/2, Aij = 1/6 off-diagonal
    pi_init = np.ones(K) / K
    A = np.ones((K, K)) / (2*(K-1)) + np.eye(K)*(0.5 - 1/(2*(K-1)))
    print("With these naive pi and A, compute p(q_t | u1,..,uT) (filtering) with T=500 and t:1,100...")
    # Compute alpha_t and beta_t from all of the 500 test points
    log_alpha, log_beta = alphabeta(u, A, mus, Sigmas, pi_init)
    # p(q_t | u1,..,uT) (filtering) with T=500 and t:1,100
    filtering = compute_filtering(log_alpha, log_beta, K, t_max)
    filtering, permutation = permutation_invariance(filtering)
    # Plot : 4 subplots, for 100 first points
    # There is a permutation of the plots each time we run the algo
    # because it is unsupervised clustering: the index is not important!
    plt.figure()
    # Fixed: added a different symbol for each cluster!
    colors = ['k', 'r', 'y', 'c']  # TODO: change from K=4 to generic K
    markers = ['o', 's', '*', 'd']  # TODO: change from K=4 to generic K
    for q in range(K):
        cq = colors[q]
        mq = markers[q]
        plt.subplot(K, 1, q+1)
        plt.plot(filtering[:, q], color=cq, marker=mq, linestyle='-')
        # plt.plot(filtering[:, q], 'b+-')
        plt.xlim([0, t_max+1])
        plt.xlabel("Time $t$ (ie. index in the data)")
        plt.ylim([0, 1.1])
        plt.ylabel(r"$\mathbb{P}(q_t = %i | u_{1..%i})$" % (q+1, t_max))
        if q == 0:
            plt.title(r"(Q2) $\alpha,\beta$-filtering (naive $\pi,A$, with $\mu_k,\Sigma_k$ learned with EM for HMM).   For state $q=1$.")
        else:
            plt.title("For state $q = {}$".format(q+1))
    print("Plotting the filtering p(q_t|u_1,..,u_T) for each state (q=1,..,{}) for the first {} points".format(K, t_max))
    if save:
        name = "question2__tmax_{}".format(t_max)
        print("Saving this figure to: fig/{}.png ...".format(name))
        plt.savefig(ospathjoin("fig", name + ".png"))
        if saveAlsoPDF:
            print("Saving this figure to: fig/{}.png ...".format(name))
            plt.savefig(ospathjoin("fig", name + ".pdf"))
    return filtering,permutation


# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# %% 3. Formal proof of the estimation / maximization equations for the EM algorithm for HMM (cf. written report)
print("\n\n3. Formal proof of the estimation / maximization equations for the EM algorithm for HMM (cf. written report) :")
print("Reference :\n - Lecture notes from the PGM course (lecture 7),\n - Bishop, Pattern Recognition and Machine Learning (§ 13.2.3 page 620).")


# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# %% 4. Implementation of the EM algorithm for HMM
print("\n\n4. Implementation of the EM algorithm for HMM :")
print("Reference :\n - Bishop, Pattern Recognition and Machine Learning (§ 13.2.3 page 620).")


def E_step_HMM(u, A, mus, Sigmas, pi_init):
    """ E step of the EM algorithm for the HMM, see our written report for the proof.

    - Return: p_Zt, p_Zt_Zt_plus_1, llh,
    - Complexity: O(T K^2),
    - Reference: [Bach15] §4.1 (page 3).
    """
    T = np.shape(u)[0]   # Number of points (time horizon)
    K = np.shape(mus)[0]  # Number of clusters
    log_alpha, log_beta = alphabeta(u, A, mus, Sigmas, pi_init)
    # Initialization, with a filtering in fact!
    p_Zt = np.zeros((T, K))
    for t in range(T):  # O(T)
        log_al_be = log_alpha[t, :] + log_beta[t, :]
        ai = log_al_be
        max_ai = np.max(ai)
        # log trick for this computation, even if it is p_Zt and not log_P_Zt !
        p_Zt[t, :] = np.exp(log_al_be - max_ai - logsumexp(ai - max_ai))
    # Loop
    p_Zt_Zt_plus_1 = np.zeros((T-1, K, K))
    for t in range(T-1):  # O(T)
        log_al_be = log_alpha[t, :] + log_beta[t, :]
        # log trick for this computation
        ai = log_al_be
        max_ai = np.max(ai)
        log_p = max_ai + logsumexp(ai - max_ai)

        u_t_plus_1 = u[t+1, :]
        for q in range(K):  # O(K)
            for q_t_plus_1 in range(K):  # O(K)
                log_p_q_t_plus_1_knowing_q_t = np.log(A[q, q_t_plus_1])

                mu_q_plus_1 = mus[q_t_plus_1, :]
                sigma_q_plus_1 = Sigmas[q_t_plus_1]
                log_p_u_t_plus_1_knowing_q_t_plus_1 = log_gaussian_density(u_t_plus_1, mu_q_plus_1, sigma_q_plus_1)

                # log trick for this computation, even if it is p_Zt and not log_P_Zt !
                p_Zt_Zt_plus_1[t, q, q_t_plus_1] = np.exp(- log_p +
                    log_alpha[t, q] +
                    log_beta[t+1, q_t_plus_1] +
                    log_p_q_t_plus_1_knowing_q_t + log_p_u_t_plus_1_knowing_q_t_plus_1)
    # Conclude with bonus computation of the log-likelihood
    llh = loglikelihood(u, A, mus, Sigmas, pi_init, T)
    return p_Zt, p_Zt_Zt_plus_1, llh


def M_step_HMM(u, A, mus, Sigmas, pi_init, p_Zt, p_Zt_Zt_plus_1):
    """ M step of the EM algorithm for the HMM, see our written report for the proof.

    - Return: (A, mus, Sigmas, pi_init) the next state of the parameters,
    - Complexity: O(T K^2),
    - Reference: Bishop, Pattern Recognition and Machine Learning, § 13.2.1 (page 617-618).
    """
    T, d = np.shape(u)    # Number of points (time horizon) and dimension
    K = np.shape(mus)[0]  # Number of clusters
    # M step
    for q in range(K):  # O(K)
        # M.1 For the q-th l-th transition probability A[q,l], cf. [Bishop (13.19)]
        for l in range(K):  # O(K)
            A[q, l] = np.sum(p_Zt_Zt_plus_1[:, q, l]) / \
                      np.sum(p_Zt_Zt_plus_1[:, q, :])  # The matrix is stochastic!
            # Remark: Bishop (13.19) sums for N from 1 and not 0!
            # XXX It is just a convention, HMM starts with N=1 in Bishop
            # Same difference between Matlab a(1:end) and Python a[0:]

        # M.2 For the q-th center mus[q], cf. [Bishop (13.20)]
        mus[q, :] = np.dot(p_Zt[:, q], u) / np.sum(p_Zt[:, q])
        # M.3 For the q-th covariance matrix Sigmas[q], cf. [Bishop (13.21)]
        tmp = u - mus[q, :]
        Sigmas[q] = sum(p_Zt[t, q]*(np.reshape(tmp[t, :], (d, 1))).dot(np.transpose(np.reshape(tmp[t, :], (d, 1)))) for t in range(0, T)) / np.sum(p_Zt[:, q])
        # This part is super ulgy, but works well.
        # XXX The O(T) comes from here!

        # M.4 For the q-th initial probability pi_init[q], cf. [Bishop (13.18)]
        # pi_init[q] = p_Zt[0, q] / np.sum(p_Zt[0, :])  # It reduces precision
        pi_init[q] = p_Zt[0, q]
        # XXX normalization seems unnecessary: p_Zt[0,:] is already a proba distribution over {1,..,K} !
    # Ideally, nothing should be to returned:
    # (A, mus, Sigmas, pi_init) were modified *in-place*
    return (A, mus, Sigmas, pi_init)


def EM_HMM(u, A, mus, Sigmas, pi_init,
           max_steps=40, tolerance=1e-3, verb=True, u_test=None):
    """ EM algorithm for the HMM, see our written report for the proof.

    - Return: (A, mus, Sigmas, pi_init, all_llh) all_llh is a list of the successive log-likelihood values,
    - If u_test is present, all_llh_test is also returned,
    - There is another stopping criterion, based on the log-likelihood: if it stops decreasing (decreasing but not enough) we stop, and if it starts increasing, that's a bug so it raise a ValueError exception,

    - At most max_steps steps are performed,
    - Complexity: O(T K^2 max_steps) at most,
    - Remark: EM for HMM is also called the Baum-Welch algorithm,
    """
    T = np.shape(u)[0]   # Number of points (time horizon)
    K = np.shape(mus)[0]  # Number of clusters
    all_llh = []
    # XXX Comparison between u (train data) and u_test
    do_comparison_test_data = True
    if u_test is None:
        do_comparison_test_data = False
    else:
        Ttest = np.shape(u_test)[0]   # Number of test points (time horizon)
    if do_comparison_test_data:
        all_llh_test = []
    if verb:
        print("Starting the EM algorithm for the HMM, with T = {}, K = {}, and max_steps = {}, tolerance = {}.".format(T, K, max_steps, tolerance))
        print("The initial belief on A, mus, Sigmas, pi_init are:")
        print("- A : \n{},\n - mus :\n{},\n - Sigmas :\n{},\n - pi_init :\n{}".format(A, mus, Sigmas, pi_init))
    # Loop for the EM algorithm (E-step, then stopping criterion, then M-step)
    for step in range(max_steps):
        # 1. E-step, returns belief hidden Zt and hidden link Zt,Zt+1, and log-likelihood
        # 1. E-step is in O(T K)
        p_Zt, p_Zt_Zt_plus_1, llh = E_step_HMM(u, A, mus, Sigmas, pi_init)
        if verb:
            print("Step {}, current log-likelihood is {:g}.".format(step, llh))
        # One more log-likelihood
        all_llh.append(llh)
        # XXX One more log-likelihood for test data
        if do_comparison_test_data:
            llh_test = loglikelihood(u_test, A, mus, Sigmas, pi_init, Ttest)
            if verb:
                print("  Step {}, current test log-likelihood is {:g}.".format(step, llh_test))
            all_llh_test.append(llh_test)
        # 2. Stopping criterion, exactly as for EM for Mixture of Gaussian
        delta_llh = llh - all_llh[step-1]
        if (step > 0) and delta_llh < - tolerance:
            print("ERROR : in EM_HMM(..) the log-likelihood is going down! (step = {}, current log-likelihood = {}) -- we try to MAXIMIZE it!".format(step, llh))
            # FIXME when done, add this ValueError (important for debugging)
            # raise ValueError("EM_HMM(..): ERROR, the log-likelihood is going up!")
        if (step > 0) and 0 < delta_llh < tolerance:
            if verb:
                print("EM_HMM(..): the log-likelihood is still increasing, but not increasing enough (with tolerance = {}), we stop here.".format(tolerance))
            break  # We stop here the for loop (before completing all the max_steps steps)
        # 3. M-step
        # 3. M-step is in O(T K^2)
        A, mus, Sigmas, pi_init = M_step_HMM(u, A, mus, Sigmas, pi_init, p_Zt, p_Zt_Zt_plus_1)
        # Done for this step of EM (E then M), continue
    # Done for the whole EM algorithm, return learned parameters and log-likelihood vector
    if do_comparison_test_data:  # XXX we compared with u_test also (cf. question 5)
        return (A, mus, Sigmas, pi_init, all_llh, all_llh_test)
    return (A, mus, Sigmas, pi_init, all_llh)


# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# %% 5. Compute and plot log-likelihood as function of the iterations of the algorithm
print("\n\n5. Compute and plot log-likelihood as function of the iterations of the algorithm :")
print("Reference :\n - (13.42) Bishop, Pattern Recognition and Machine Learning, § 13.2.1 (page 617).")


def loglikelihood(u, A, mus, Sigmas, pi_init, n):
    r""" Compute the log-likelihood of a predicted sequence for HMM (given its parameters A, mus, Sigmas, pi_init).

    - log p(u) = log sum_{z_N} alpha(z_N) = logsumexp_{z_N} log alpha(z_N).
    - Computed with an alpha recursion (O(T K) time complexity).
    - Reference: (13.42) Bishop, Pattern Recognition and Machine Learning, § 13.2.1 (page 617).
    """
    T = np.shape(u)[0]   # Number of points (time horizon)
    log_alpha = alpha_recursion(u, A, mus, Sigmas, pi_init)
    last_log_alpha = log_alpha[T-1, :]
    # log-sum-exp trick !
    max_log_alpha = np.max(last_log_alpha)
    llh = max_log_alpha + logsumexp(last_log_alpha - max_log_alpha)
    # XXX as done in the correction of HMK 2, we divide the log-likelihood by the number of training/testing points
    # To be able to compare it better, if Ntrain and Ntest are different.
    return llh / n


def likelihood(u, A, mus, Sigmas, pi_init, n):
    """ Compute the likelihood of a predicted sequence for HMM (given its parameters A, mus, Sigmas, pi_init)."""
    return np.exp(n * loglikelihood(u, A, mus, Sigmas, pi_init, n))


def question5(u, u_test, K=4, max_iter=-1, init=None,
        title=None, save=doSaveThePlots):
    """ Question 5 : Plot the log-likelihood evolving during the EM algorithm for the train and the test data.

    - max_iter: max iteration of the preliminar EM for Mixture of Gaussian, can be -1 only initilize with K-means++ (and a converged K-means),
    - init: None to use K-means then EM (or juste K-means), or 'random' for STUPID initialization (random centroids, BIG std),
    - title: to change the title of the plot (to compare the three initialization methods),
    - save: True to save the plots.
    """
    n, d = np.shape(u)
    if init == 'random':
        print("Using STUPID random initialization (no Kmeans, no MoG).")
        inds = np.random.choice(range(n), K, replace=False)
        mus = u[inds]
        # Initial variances cluster-wise
        Sigmas = np.zeros((K, d, d))  # now we do not have Sj = sj * I_d
        for j in range(K):
            Sigmas[j] = np.cov(u.T)
    else:
        print("Using the means and covariances obtained from MoG from DM2.")
        mus, Sigmas, _ = EM_general(u, K, max_iter=max_iter)
    # Initial pi (uniform), A estimations :
    # Aii = 1/2, Aij = 1/6 off-diagonal
    pi_init = np.ones(K) / K
    A = np.ones((K, K)) / (2*(K-1)) + np.eye(K)*(0.5 - 1/(2*(K-1)))
    # Learn the HMM parameters from all of the 500 train points
    A, mus, Sigmas, pi_init, all_llh, all_llh_test = EM_HMM(u, A, mus, Sigmas, pi_init, u_test=u_test)
    print("EM for HMM is done, we learned these parameters:")
    print(" - Learned centers (mus) are:\n", mus)
    print(" - Learned Sigmas are:\n", Sigmas)
    print(" - Learned transition matrix A is:\n", A)
    print(" - Learned initial probability pi_init is:\n", pi_init)
    # Plot the log-likelihood
    plt.figure()
    plt.plot(all_llh, 'r*-')
    plt.plot(all_llh_test, 'b*-')
    plt.legend([
            "Train log-likelihood",
            "Test log-likelihood"
        ], loc='lower right')
    plt.xlabel("Step in the EM algorithm (for HMM)")
    plt.ylabel("Log-likelihood (log-scale)")
    if title is not None:
        title = "\n" + title
    else:
        title = ""
    plt.title("(Q5) Train and test log-likelihood as a function of EM steps (HMM K={} states){}".format(K, title))
    if save:
        name = "question5__{}_{}".format(init, max_iter)
        print("Saving this figure to: fig/{}.png ...".format(name))
        plt.savefig(ospathjoin("fig", name + ".png"))
        if saveAlsoPDF:
            print("Saving this figure to: fig/{}.png ...".format(name))
            plt.savefig(ospathjoin("fig", name + ".pdf"))
    # Done
    return A, mus, Sigmas, pi_init, all_llh, all_llh_test


# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# %% 6. Compute and print log-likelihood of the Mixture of Gaussian and HMM models (on train and test)
print("\n\n6. Compute and print log-likelihood of the Mixture of Gaussian and HMM models (on train and test) :")


def question6(Xtrain, Xtest, n, d, K=4, max_iter=-1, init=None):
    """ Question 6 : Display the log-likelihood for both MoG and HMM models, for the train and the test data."""
    # 1) For Mixture of Gaussians
    # For Xtrain
    print("Training Mixture of Gaussians on the train data...")
    mus_MoG, Sigmas_MoG, weights_MoG = EM_general(Xtrain, K)
    print("Predicting clusters (with Mixture of Gaussians) on the train data...")
    labels_train_MoG = EM.predict_general(Xtrain, mus_MoG, Sigmas_MoG, weights_MoG, n, d, K)
    clusters_train_MoG = EM.build_clusters_general(Xtrain, labels_train_MoG, n, K)
    print("Computing log-likelihood (with Mixture of Gaussians) on the train data...")
    llh_train_MoG = EM.loglikelihood_MoG(clusters_train_MoG, mus_MoG, Sigmas_MoG, weights_MoG, K, n)
    print("llh_train_MoG =", llh_train_MoG)
    # For Xtest
    print("Predicting clusters (with Mixture of Gaussians) on the test data...")
    labels_test_MoG = EM.predict_general(Xtest, mus_MoG, Sigmas_MoG, weights_MoG, n, d, K)
    clusters_test_MoG = EM.build_clusters_general(Xtest, labels_test_MoG, n, K)
    print("Computing log-likelihood (with Mixture of Gaussians) on the test data...")
    llh_test_MoG = EM.loglikelihood_MoG(clusters_test_MoG, mus_MoG, Sigmas_MoG, weights_MoG, K, n)
    print("llh_test_MoG =", llh_test_MoG)
    # 2) For HMM
    A, mus, Sigmas, pi_init, all_llh, all_llh_test = question5(Xtrain, Xtest, K=4,
        max_iter=max_iter, init=init, save=False)
    # For Xtrain
    llh_train_HMM = all_llh[-1]
    print("llh_train_HMM =", llh_train_HMM)
    # For Xtest
    llh_test_HMM = all_llh_test[-1]
    print("llh_test_HMM =", llh_test_HMM)
    # Printing a LaTeX code for this table
    # XXX use https://pypi.python.org/pypi/tabulate for a more general printing solution
    table = r"""%% -*- mode: latex; coding: utf-8 -*-
%% Generated with HMM.py (function question6)
    Table of the 4 normalized log-likelihoods (asked for question 6):
    \begin{center}
        \begin{tabular}{|l|c|c|}
            \hline
            Model & Train log-likelihood & Test log-likelihood \\
            \hline
            \hline
            Mixture of Gaussians (MoG) & $%g$ & $%g$ \\
            \hline
            Hidden Markov Model (HMM) & $%g$ & $%g$ \\
            \hline
            Hidden semi-Markov Model (HSMM) & \emph{not yet} & \emph{not yet} \\
            \hline
        \end{tabular}
    \end{center}
%% End of question6.tex
""" % (llh_train_MoG, llh_test_MoG, llh_train_HMM, llh_test_HMM)
    print(table)
    with open("question6.tex", "w") as out_table:
        out_table.write(table)
    return (llh_train_MoG, llh_test_MoG, llh_train_HMM, llh_test_HMM)


# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# %% 7. Description and pseudo-code for the Viterbi algorithm (cf. written report)
print("\n\n7. Description and pseudo-code for the Viterbi algorithm (cf. written report) :")
print("Reference :\n - https://en.wikipedia.org/wiki/Viterbi_algorithm\n - Bishop, Pattern Recognition and Machine Learning, § 13.2.3 (page 623).")


# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# %% 8. Implementation of the Viterbi algorithm (aka MAP inference, or max-product algorithm)
print("\n\n8. Implementation of the Viterbi algorithm (aka MAP inference, or max-product algorithm) :")
print("Reference :\n - https://en.wikipedia.org/wiki/Viterbi_algorithm\n - Bishop, Pattern Recognition and Machine Learning, § 13.2.5 (page 629).")


def viterbi(u, A, mus, Sigmas, pi_init):
    """ Viterbi algorithm for the HMM, see our written report for the pseudo-code and explanations.

    - Return: viterbi_path is the path, which can be retrieved by saving back pointers that remember which state x was used in the second equation,
    - Remark: Viterbi for HMM is also called MAP inference, or max-product algorithm),
    - Reference:
      - Bishop, Pattern Recognition and Machine Learning, § 13.2.5 (page 629),
      - https://en.wikipedia.org/wiki/Viterbi_algorithm#Algorithm.
    """
    T = np.shape(u)[0]   # Number of points (time horizon)
    K = np.shape(mus)[0]  # Number of clusters
    # Initialization
    omegas = np.zeros((T, K))
    indeces = np.zeros((T, K))
    # 1. Initialization for omegas (Eq. 13.69, [Bishop, page 630])
    for q in range(K):
        mu_q = mus[q, :]
        sigma_q = Sigmas[q]
        log_p_u_t_knowing_q_t = log_gaussian_density(u[0, :], mu_q, sigma_q)
        # Virtual message for alpha[0]
        omegas[0, q] = log_p_u_t_knowing_q_t + np.log(pi_init[q])
    # 2. Loop (Eq. 13.68, [Bishop, page 630])
    for t in range(1, T):  # forward !
        ut = u[t, :]
        for q in range(K):
            mu_q = mus[q, :]
            sigma_q = Sigmas[q]
            # From the Gaussian density for state q : mu_q, sigma_q
            log_p_u_t_knowing_q_t = log_gaussian_density(ut, mu_q, sigma_q)
            # Here we use the definition of the max-product algorithm
            # but as we use log, it's a max-sum (on the logs)
            # Warning : in Octave/Matlab, max returns [ value, index ], not in Python
            log_p_u_t_knowing_q_t__a_xq__V_t_minus_1 = log_p_u_t_knowing_q_t + omegas[t-1, :] + np.log(A[:, q].T)
            om = np.max(log_p_u_t_knowing_q_t__a_xq__V_t_minus_1)
            omegas[t, q] = om
            idx = np.argmax(log_p_u_t_knowing_q_t__a_xq__V_t_minus_1)
            indeces[t, q] = idx
    # Build the path, from these indeces, coming from the end
    viterbi_path = np.zeros(T)
    last_index = np.argmax(omegas[T-1, :])
    viterbi_path[T-1] = last_index
    for t in range(T-1, 0, -1):  # Cf. [Bishop, § 13.2.5], (13.71)
        viterbi_path[t-1] = indeces[t, viterbi_path[t]]
    # We are done, returning the path.
    return viterbi_path


def predict_HMM(u, A, mus, Sigmas, pi_init):
    """ Use the learned parameters A, mus, Sigmas, pi_init to predict the values of the hidden state (ie predict the clusters).

    - This is exactly what is done by the Viterbi algorithm
    - WARNING: it is NOT an argmax on the predicted values for proba of hidden values. We could pick a sequence which has zero probability. Cf. [Bishop, § 13.2.5] first paragraph.
    """
    most_likely_sequence_of_states = viterbi(u, A, mus, Sigmas, pi_init)
    # Maybe permutation_invariance could also return its permutation, given here and applied here?
    # most_likely_sequence_of_states = permutation_invariance(most_likely_sequence_of_states)
    return most_likely_sequence_of_states


print("\n\n8. Plot the clusterization from the parameters learned with EM (for HMM).")


def plot_2D_clusterization_from_EM_HMM(u, A, mus, Sigmas, pi_init, permutation, testOrTrain="train", save=doSaveThePlots):
    """ Use the Viterbi algorithm to predict the most likely sequence of (hidden) state, and plot it.

    - Plot this (predicted) hidden sequence as a 2D colorful clusterization (like K-means for the DM2).
    """
    T = np.shape(u)[0]
    K = np.shape(mus)[0]  # Number of clusters
    # First we use Viterbi to predict the most likely sequence of state
    most_likely_sequence_of_states = predict_HMM(u, A, mus, Sigmas, pi_init)
    # Convert to labels
    labels = most_likely_sequence_of_states
    labels = permutation_viterbi(labels,permutation,K,T)

    # Then we plot
    print("Plotting the 2D data ({} points) with centroids and clusters...".format(u.shape[0]))
    plt.figure()
    plt.hold(True)
    # Plot the training data colored by clusters
    # plt.scatter(u[:, 0], u[:, 1], c=labels, edgecolors='', s=60)
    # Fixed: added a different symbol for each cluster!
    colors = ['k', 'r', 'y', 'c']  # TODO: change from K=4 to generic K
    markers = ['o', 's', '*', 'd']  # TODO: change from K=4 to generic K
    for k in range(0, K):
        uk = u[labels == k]
        ck = colors[k]
        mk = markers[k]
        print("For state k = {}, using color = {} and marker = {}, with {} points.".format(k, ck, mk, np.shape(uk)[0]))
        plt.scatter(uk[:, 0], uk[:, 1], c=ck, marker=mk, edgecolors='', s=60)
    # Plot the clusters centers (big O plus)
    plt.scatter(mus[:, 0], mus[:, 1], s=400, marker='o', edgecolors='k', linewidths=2, c='none')
    plt.scatter(mus[:, 0], mus[:, 1], s=400, marker='+', edgecolors='k', linewidths=2)  # , c=range(n_clusters))
    plt.title(r"Clusters $k=1..%i$ predicted by Viterbi, with $(\pi,A,\mu_k,\Sigma_k)$ learned by EM (HMM)." % K)
    plt.hold(False)
    # Save the plots?
    if save:
        name = "question8_"+testOrTrain
        print("Saving this figure to: fig/{}.png ...".format(name))
        plt.savefig(ospathjoin("fig", name + ".png"), dpi=160)
        if saveAlsoPDF:
            print("Saving this figure to: fig/{}.pdf ...".format(name))
            plt.savefig(ospathjoin("fig", name + ".pdf"), dpi=190)
    plt.show()
    return labels


question8 = plot_2D_clusterization_from_EM_HMM


# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# %% 9. Marginal probabilities of being in state k (for first 100 test points)
print("\n\n9. Marginal probabilities of being in stage k (for first 100 test points) :")


def question9(u_test, A, mus, Sigmas, pi_init, K=4, t_max=100, save=doSaveThePlots):
    """ Question 9 : K plots, clusterization from a learned HMM."""
    print("With the learned parameters pi and A, compute p(q_t | u1,..,uT) (filtering) with T=500 and t:1,100...")
    # Compute alpha_t and beta_t from all of the 500 TEST points
    log_alpha, log_beta = alphabeta(u_test, A, mus, Sigmas, pi_init)
    # p(q_t | u1,..,uT) (filtering) with T=500 and t:1,100
    filtering = compute_filtering(log_alpha, log_beta, K, t_max)
    filtering, _ = permutation_invariance(filtering)
    # Plot : 4 subplots, for 100 first points
    # There is a permutation of the plots each time we run the algo
    # because it is unsupervised clustering: the index is not important!
    plt.figure()
    # Fixed: added a different symbol for each cluster!
    colors = ['k', 'r', 'y', 'c']  # TODO: change from K=4 to generic K
    markers = ['o', 's', '*', 'd']  # TODO: change from K=4 to generic K
    for q in range(K):
        cq = colors[q]
        mq = markers[q]
        plt.subplot(K, 1, q+1)
        plt.plot(filtering[:, q], color=cq, marker=mq, linestyle='-')
        plt.xlim([0, t_max+1])
        plt.xlabel("Time $t = 1 .. %i$ (ie. index in the data)." % t_max)
        plt.ylim([0, 1.1])
        plt.ylabel(r"$\mathbb{P}(q_t = %i | u_{1..%i})$" % (q+1, t_max))
        if q == 0:
            plt.title(r"(Q9) Complete $\alpha,\beta$-filtering (with parameters $\pi,A,\mu_k,\Sigma_k$ learned with EM for HMM).   For state $q=1$.")
        else:
            plt.title("For state $q = {}$.".format(q+1))
    print("Plotting the filtering p(q_t|u_1,..,u_T) for each state (q=1,..,{}) for the first {} points".format(K, t_max))
    if save:
        name = "question9"
        print("Saving this figure to: fig/{}.png ...".format(name))
        plt.savefig(ospathjoin("fig", name + ".png"))
        if saveAlsoPDF:
            print("Saving this figure to: fig/{}.pdf ...".format(name))
            plt.savefig(ospathjoin("fig", name + ".pdf"))
    return filtering


# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# %% 10. Plot affectation as a function of time (discrete values) (for first 100 test points)
print("\n\n10. Plot affectation as a function of time (discrete values) (for first 100 test points) :")


def question10(u_test, A, mus, Sigmas, pi_init, K=4, t_max=100, save=doSaveThePlots):
    """ Question 10 : Plot affectation as a function of time (discrete values) (for first 100 test points)."""
    print("With the learned parameters pi and A, compute p(q_t | u1,..,uT) (filtering) with T=500 and t:1,100...")
    # Compute alpha_t and beta_t from all of the 500 TEST points
    log_alpha, log_beta = alphabeta(u_test, A, mus, Sigmas, pi_init)
    # p(q_t | u1,..,uT) (filtering) with T=500 and t:1,100
    filtering = compute_filtering(log_alpha, log_beta, K, t_max)
    filtering, _ = permutation_invariance(filtering)

    print("Then predicting the most likely state, one by one, with an argmax.")
    print("WARNING: Note that this is not the most likely sequence of state !")
    affected_state = np.argmax(filtering, axis=1)
    # Warning this is not the good way to predict the most likely sequence of hidden variables, cf. explanations about the Viterbi algorithm.

    plt.figure()
    plt.plot(1 + affected_state, 'b*', markersize=15)
    plt.xlim([0, t_max+1])
    plt.xlabel("Time $t = 1 .. %i$ (ie. index in the data)." % t_max)
    plt.ylim([0.5, K+0.5])
    plt.yticks(list(range(1, K+1)))
    plt.ylabel(r"Most likely affected state $q \in \{1,..,%i\}$" % K)
    plt.title(r"(Q10) Most likely states $q \in \{1,..,%i\}$ according to marginal probability ($\mathrm{argmax}$, not Viterbi)" % K)
    if save:
        name = "question10"
        print("Saving this figure to: fig/{}.png ...".format(name))
        plt.savefig(ospathjoin("fig", name + ".png"))
        if saveAlsoPDF:
            print("Saving this figure to: fig/{}.pdf ...".format(name))
            plt.savefig(ospathjoin("fig", name + ".pdf"))
    return affected_state


# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# %% 11. Viterbi on the test data, compare with the previous clusterization
print("\n\n11. Viterbi on the test data, compare with the previous clusterization :")


def question11(u, u_test, A, mus, Sigmas, pi_init, K=4, t_max=100, save=doSaveThePlots):
    """ Question 11 : Plot affectation as a function of time (discrete values) (for first 100 test points)."""
    print("With the learned parameters pi and A, compute p(q_t | u1,..,uT) (filtering) with T=500 and t:1,100...")
    print("Then predicting the most likely state sequence of state, with Viterbi algorithm")
    affected_state = predict_HMM(u, A, mus, Sigmas, pi_init)
    # This is the good way to predict the most likely sequence of hidden variables, cf. explanations about the Viterbi algorithm.
    plt.figure()
    plt.plot(1 + affected_state, 'r*', markersize=15)
    plt.xlim([0, t_max+1])
    plt.xlabel("Time $t = 1 .. %i$ (ie. index in the data)." % t_max)
    plt.ylim([0.5, K+0.5])
    plt.yticks(list(range(1, K+1)))
    plt.ylabel(r"Most likely affected state $q \in \{1,..,%i\}$" % K)
    plt.title(r"(Q11) Most likely states  $q \in \{1,..,%i\}$ according to marginal probability, predicted by Viterbi algorithm." % K)
    if save:
        name = "question11"
        print("Saving this figure to: fig/{}.png ...".format(name))
        plt.savefig(ospathjoin("fig", name + ".png"))
        if saveAlsoPDF:
            print("Saving this figure to: fig/{}.pdf ...".format(name))
            plt.savefig(ospathjoin("fig", name + ".pdf"))
    return affected_state

def permutation_viterbi(viterbi_path,permutation,K,T):
    new_viterbi_path = np.zeros(T)
    for t in range(T):
        new_viterbi_path[t] = permutation[viterbi_path[t]]
    return new_viterbi_path

def question10_11(u, u_test, A, mus, Sigmas, pi_init, K=4, t_max=100, save=doSaveThePlots):
    T = np.shape(u)[0]
    # Max of marginals:
    log_alpha, log_beta = alphabeta(u_test, A, mus, Sigmas, pi_init)
    filtering = compute_filtering(log_alpha, log_beta, K, t_max)
    filtering, permutation = permutation_invariance(filtering)
    affected_state = np.argmax(filtering, axis=1)

    plt.subplot(2, 1, 1)
    plt.plot(1 + affected_state, 'b')
    plt.xlim([0, t_max+1])
    plt.ylim([0.5, K+0.5])
    plt.yticks(list(range(1, K+1)))
    plt.ylabel("state")
    plt.xlabel("max of marginals")

    # Viterbi:
    affected_state_viterbi = predict_HMM(u, A, mus, Sigmas, pi_init)
    affected_state_viterbi = permutation_viterbi(affected_state_viterbi,permutation,K,T)

    plt.subplot(2, 1, 2)
    plt.plot(1 + affected_state_viterbi, 'r')
    plt.xlim([0, t_max+1])
    plt.ylim([0.5, K+0.5])
    plt.yticks(list(range(1, K+1)))
    plt.ylabel('state')
    plt.xlabel('Viterbi sequence')
    plt.tight_layout()


    if save:
        name = "questions1011_sameGraph"
        print("Saving this figure to: fig/{}.png ...".format(name))
        plt.savefig(ospathjoin("fig", name + ".png"))
        if saveAlsoPDF:
            print("Saving this figure to: fig/{}.pdf ...".format(name))
            plt.savefig(ospathjoin("fig", name + ".pdf"))
    return affected_state,affected_state_viterbi


# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# %% 12. Experimental guess of the best number of states
print("\n\n12. Experimental guess of the best number of states :")


def guess_best_K(u, u_test, Kmax=10,
        max_iter=100, init=None, title=None, save=doSaveThePlots):
    """ Experimental guess of the best number of states.

    - Plot the train and test log-likelihoods as function of possible value of K (2 <= K <= Kmax), and decide the best one.
    - See the previous function question5 for documentation on the arguments to give here,
    - Return: best_K, best_A, best_mus, best_Sigmas, best_pi_init, best_llh, best_llh_test, the best value for K = argmax train log-likelihood(k), and the parameters for this value of K.
    - Warning: it is very time consuming! For each K it performs a lot of operations (K-means, EM for MoG, and then EM for HMM)...
    - Warning: best K is only decided on llh_train, so THERE IS OVERFITTING! WARNING! Decide based on the plot, not on the max llh_train !!

    - Remark: I tried it, for train/test data given in the DM3, with a small value for Kmax = 10. See the plot fig/question12.png attached with the DM. Or https://bitbucket.org/lbesson/mva15-project-probabilistic-graphical-models/commits/15c755d.
    """
    # Initialization, getting sizes etc
    T_train = np.shape(u)[0]  # Number of train points (time horizon)
    T_test = np.shape(u_test)[0]  # Number of test points (time horizon)
    n, d = np.shape(u)
    if Kmax is None:
        Kmax = min(T_train, T_test)
    Kmax = max(2, min(Kmax, min(T_train, T_test)))
    print("Testing all K for K = 2 .. Kmax (Kmax = {}).".format(Kmax))
    all_K = list(range(1, Kmax+1))
    # Now the llh are stored for each K (not each t in the EM algo!)
    all_llh = np.zeros((Kmax, 1))
    all_llh_test = np.zeros((Kmax, 1))
    # Best parameters
    best_K = None
    best_llh_train = -np.inf
    # Starting the big loop on possible nb of cluster K
    for K in all_K:
        print("# For K = {}...".format(K))
        # Getting initial guess for mus, Sigmas
        if init == 'random':
            print("    Using STUPID random initialization (no Kmeans, no MoG)...")
            inds = np.random.choice(range(n), K, replace=False)
            mus = u[inds]
            # Initial variances cluster-wise
            Sigmas = np.zeros((K, d, d))  # now we do not have Sj = sj * I_d
            for j in range(K):
                Sigmas[j] = np.cov(u.T)
        else:
            print("    Using the means and covariances obtained from MoG from DM2...")
            mus, Sigmas, _ = EM_general(u, K, max_iter=max_iter)
        # Initial pi (uniform), A estimations :
        print("    Initial uniform pi and A...")
        # Aii = 1/2, Aij = 1/6 off-diagonal
        pi_init = np.ones(K) / K
        if K > 1:
            A = np.ones((K, K)) / (2*(K-1)) + np.eye(K)*(0.5 - 1/(2*(K-1)))
        else:
            A = np.ones((K, K))
        # Learn the HMM parameters from all of the 500 train points
        print("    Starting to learn the HMM parameters with EM...")
        A, mus, Sigmas, pi_init, all_llh_k_train, all_llh_k_test = EM_HMM(u, A, mus, Sigmas, pi_init, u_test=u_test)
        print("     - Learned centers (mus) are:\n", mus)
        print("     - Learned Sigmas are:\n", Sigmas)
        print("     - Learned transition matrix A is:\n", A)
        print("     - Learned initial probability pi_init is:\n", pi_init)
        # llh_train, llh_test = function(u, u_test, A, mus, Sigmas, pi_init, K=K)
        llh_train = all_llh_k_train[-1]
        all_llh[K-2] = llh_train
        print("     - Train log-likelihood =", llh_train)
        llh_test = all_llh_k_test[-1]
        all_llh_test[K-2] = llh_test
        print("     - Test log-likelihood =", llh_test)
        if llh_train > best_llh_train:
            best_llh_train = llh_train
            best_K = K
            print("New best choice for K = ", best_K)
            best_A, best_mus, best_Sigmas, best_pi_init = A, mus, Sigmas, pi_init
            best_llh, best_llh_test = llh_train, llh_test
    # Plot the train/test log-likelihood as functions of K
    plt.figure()
    plt.plot(all_K, all_llh, 'r*-')
    plt.plot(all_K, all_llh_test, 'b*-')
    plt.legend([
            "Train log-likelihood",
            "Test log-likelihood"
        ], loc='lower right')
    plt.xlabel("Number of clusters $K = 1..{}$".format(Kmax))
    plt.xlim([0.5, Kmax+1.5])
    plt.ylabel("Normalized log-likelihood (log-scale)")
    if title is not None:
        title = "\n" + title
    else:
        title = ""
    plt.title("(Q12) Train and test log-likelihood as a function of nb clusters (HMM){}".format(title))
    plt.show()
    if save:
        name = "question12"
        print("Saving this figure to: fig/{}.png ...".format(name))
        plt.savefig(ospathjoin("fig", name + ".png"))
        if saveAlsoPDF:
            print("Saving this figure to: fig/{}.pdf ...".format(name))
            plt.savefig(ospathjoin("fig", name + ".pdf"))
    # Now compute the best K, and its parameters
    return best_K, best_A, best_mus, best_Sigmas, best_pi_init, best_llh, best_llh_test
    # Done for guess_best_K


question12 = guess_best_K


print("\nDone for HMM.py")
# End of HMM.py
